-- ILLiad Reshare

require("AtlasHelpers");
require("JsonParser");
require("JsonWriter");
require("Utilities");
require("TagProcessor");

luanet.load_assembly("System");
luanet.load_assembly("System.Data");
luanet.load_assembly("log4net");

local types = {};
types["System.Type"] = luanet.import_type("System.Type");
types["System.String"] = luanet.import_type("System.String");
types["System.Array"] = luanet.import_type("System.Array");
types["System.Activator"] = luanet.import_type("System.Activator");
types["System.Reflection.Assembly"] = luanet.import_type("System.Reflection.Assembly");
types["SqlDbType"] = luanet.import_type("System.Data.SqlDbType");
types["System.DateTimeOffset"] = luanet.import_type("System.DateTimeOffset");

-- Import the LogManager type
types["log4net.LogManager"] = luanet.import_type("log4net.LogManager");

local reshareAssembly = types["System.Reflection.Assembly"].LoadFile(AddonInfo.Directory .. "\\AtlasSystems.Reshare.dll");
types["AtlasSystems.Reshare.ReshareRequestClient"] = reshareAssembly:GetType("AtlasSystems.Reshare.ReshareRequestClient");
types["AtlasSystems.Reshare.ReshareSearchClient"] = reshareAssembly:GetType("AtlasSystems.Reshare.ReshareSearchClient");
types["AtlasSystems.Reshare.Models.ReshareRequest"] = reshareAssembly:GetType("AtlasSystems.Reshare.Models.ReshareRequest");
types["AtlasSystems.Reshare.Models.ReshareSearchRequestList"] = reshareAssembly:GetType("AtlasSystems.Reshare.Models.ReshareSearchRequestList");


local Settings = {};
Settings.Nvtgc = GetSetting("NVTGC");
Settings.BaseUrl = GetSetting("ReshareAPIUrl")
Settings.Tenant = GetSetting("ReshareTenant");
Settings.OpenUrlListener = GetSetting("OpenURLListenerURL");
Settings.OpenUrlTenant = GetSetting("OpenURLTenant");
Settings.Username = GetSetting("ReshareAPIUsername");
Settings.Password = GetSetting("ReshareAPIPassword");
Settings.SharedIndexUrl = GetSetting("SharedIndexAPIUrl");
Settings.SearchMonitorQueue = GetSetting("SearchingMonitorQueue");
Settings.NotFoundQueue = GetSetting("NotFoundQueue");
Settings.UnfilledQueue = GetSetting("UnfilledQueue");
Settings.RequestSentQueue = GetSetting("RequestSentQueue");
Settings.ErrorQueue = GetSetting("ErrorQueue");
Settings.ProcessIdleRequests = GetSetting("SendIdleRequestsToErrorQueue");
Settings.PollingInterval = GetSetting("RequestPollingFrequency");
Settings.FilledStateCodes = GetSetting("ReshareFilledStatuses");
Settings.UnfilledStateCodes = GetSetting("ReshareUnfilledStatuses");
Settings.PatronIdField = GetSetting("PatronIDField");
Settings.PickupLocationLookupValue = GetSetting("PickupLocationLookupValue");
Settings.PickupLocationLookupMapping = {};
Settings.RequestPlaceholderId = GetSetting("RequestPlaceholderId");
Settings.NoteAndLogPrefix = GetSetting("NoteAndLogPrefix");

-- Create a logger
local log = types["log4net.LogManager"].GetLogger(Settings.NoteAndLogPrefix);

-- Make sure the placeholder setting has a value.
if (Settings.RequestPlaceholderId == nil or Settings.RequestPlaceholderId == "") then
	Settings.RequestPlaceholderId = "ReShareIdPending";
end

local requestClient;
local searchClient;

local isCurrentlyProcessing = false;
local lastPolledTime = 0;

-- These are used by the polling process.
-- hridList is a list of ReShare HRIDs that have pending updates,
-- updateParams is an HRID-indexes table of the new status, lender, and audit message for each.
local hridList = {};
local updateParams = {};

function Init()
	local loadParams = types["System.Array"].CreateInstance(types["System.Type"].GetType("System.Object"), 1);
	loadParams[0] = AddonInfo.Directory;

	local reshareLoadMethod = types["AtlasSystems.Reshare.ReshareRequestClient"]:GetMethod("LoadDependencies");
	reshareLoadMethod:Invoke(nil, loadParams);

	local requestClientParams = { Settings.BaseUrl, Settings.OpenUrlListener, Settings.Tenant, Settings.OpenUrlTenant, Settings.Username, Settings.Password };
	requestClient = types["System.Activator"].CreateInstance(types["AtlasSystems.Reshare.ReshareRequestClient"], requestClientParams);

	log:Debug("Creating ReshareSearchClient.");
	searchClient = types["System.Activator"].CreateInstance(types["AtlasSystems.Reshare.ReshareSearchClient"], { Settings.SharedIndexUrl });

	BuildPickupLocationMappings();

	log:Debug("Clients created")

    RegisterSystemEventHandler("GenerateFulfillmentData","GenerateFulfillmentData");
	RegisterSystemEventHandler("AutomationDecision","AutomationDecision");
    RegisterSystemEventHandler("SystemTimerElapsed","TimerElapsed");
end

function TimerElapsed(eventargs)
	log:Info("Reshare: Timer elapsed.")
	if (isCurrentlyProcessing == false) then
		isCurrentlyProcessing = true;
		local success, err = pcall(function()
			ProcessDataContexts("TransactionStatus", Settings.SearchMonitorQueue, "SearchAndRequest");

			if (os.difftime(os.time(), lastPolledTime) / 60 >= Settings.PollingInterval) then
				log:Info("Polling interval has elapsed.")
				lastPolledTime = os.time();
				local success, err = pcall(CheckReshareStatusUpdates);
				if (success == false) then
					log:Error("Error encountered polling Reshare requests by status.");
					log:Error(err.Message);
				end

			else
				log:Debug("Polling interval has not elapsed.")
			end
		end);

		if not success then
			log:Error("Error encountered processing ReShare requests.");
			log:Error(err.Message)
		end

		isCurrentlyProcessing = false;
	end
end

-- Stage 2
function GenerateFulfillmentData(eventArgs)
	log:Info("Reshare: Generating Fulfillment Data");

	local success, result = pcall(GetItemsAvailableInSharedIndex);

	if (not success) then
		local errorMessage = "An error occurred when searching Reshare for tn " .. GetFieldValue("Transaction", "TransactionNumber") .. ": " .. tostring(TraverseError(result));
		AddNote(GetFieldValue("Transaction", "TransactionNumber"), errorMessage);
		error("An error occurred searching Reshare");
	end
	local idList = result;

	-- idList is only nil when a bib search returns more than one result.
	if idList == nil then
		eventArgs.Available = false;
		eventArgs.Notes = "More than one result when searching with bib info. Please add a standard number to request for better results.";
	elseif idList.Count == 0 then
		log:Debug("No results found");
		eventArgs.Available = false;
		eventArgs.Notes = "No loanable holdings found in shared index.";
	else
		log:Debug(idList.Count.." results found.");
		eventArgs.Notes = idList.Count .. " records found.";
		eventArgs.Available = true;
		eventArgs.Cost = true;
		eventArgs.CopyrightIncluded = true;
		eventArgs.TurnaroundTime = 0

		local customData = {};
		customData.instanceId = idList[0].Id;
		customData.title = idList[0].Title;
		eventArgs.CustomData = JsonWriter:Serialize(customData)
	end

    eventArgs.Handled = true;
end

-- Stage 4
function AutomationDecision(eventArgs)
	log:Info("Reshare: Submitting Request from DSP");

	local customData = JsonParser:ParseJSON(eventArgs.CustomData);
	local instanceId = customData.instanceId;
	local title = customData.title;
	local patronId = TagProcessor:ReplaceTags(Settings.PatronIdField);
	local pickupLocationValue = TagProcessor:ReplaceTags(Settings.PickupLocationLookupValue);

	if pickupLocationValue == nil then
		pickupLocationValue = "";
	end

	if ((Settings.PickupLocationLookupMapping[pickupLocationValue] ~= nil) and (Settings.PickupLocationLookupMapping[pickupLocationValue] ~= "")) then
		pickupLocationValue = Settings.PickupLocationLookupMapping[pickupLocationValue];
	else
		log:WarnFormat("PickupLocation mapping not found for {0}", pickupLocationValue);
	end

	local success, err = pcall(CreateRequest, patronId, pickupLocationValue, instanceId, title, true);
	if success then
		eventArgs.Handled = true;
	else
		local tn = GetFieldValue("Transaction", "TransactionNumber");
		local errorMessage = "An error occurred when creating a ReShare request for TN " .. tn .. ": " .. tostring(TraverseError(err));

		log:Debug(errorMessage);
		AddNote(tn, errorMessage);
		eventArgs.Handled = false;
	end
end

function SearchAndRequest()
	local success, err = pcall(function()
		--Confirm that this is a request we should be processing
		local userNvtgc = GetFieldValue("User", "NVTGC");
		if (string.lower(ILLiad.GetParentNVTGC(userNvtgc)) ~= string.lower(Settings.Nvtgc)  ) then
			log:Debug("Skipping shared index search for transaction " .. GetFieldValue("Transaction", "TransactionNumber") .. " because it is not within this addon's NVTGC.");
			return;
		end

		log:Debug("Searching shared index for tn "..GetFieldValue("Transaction", "TransactionNumber"));

		local idList = GetItemsAvailableInSharedIndex();
		
		if idList then
			for i = 0, idList.Count -1 do
				log:Debug("Result " .. i);
				log:Debug("ID: " .. idList[i].Id);
				log:Debug("Title: " .. idList[i].Title);
			end
		end
		
		-- idList is only nil when a bib search returns more than one result.
		if (idList == nil or idList.Count == 0) then
			local noteText;
			if idList and idList.Count == 0 then
				noteText = "No results found. Cannot create ReShare request.";
			else
				noteText = "More than one result when searching with bib info. Please add a standard number to request for better results.";
			end

			AddNote(GetFieldValue("Transaction", "TransactionNumber"), noteText);
			ExecuteCommand("Route", {GetFieldValue("Transaction","TransactionNumber"), Settings.NotFoundQueue});
		elseif (idList.Count > 0) then
			log:Debug(idList.Count .. " result(s) found for tn " .. GetFieldValue("Transaction", "TransactionNumber"));
			local patronId = TagProcessor:ReplaceTags(Settings.PatronIdField);

			local pickupLocationValue = TagProcessor:ReplaceTags(Settings.PickupLocationLookupValue);
			if pickupLocationValue == nil then
				pickupLocationValue = "";
			end

			if ((Settings.PickupLocationLookupMapping[pickupLocationValue] ~= nil) and (Settings.PickupLocationLookupMapping[pickupLocationValue] ~= "")) then
				pickupLocationValue = Settings.PickupLocationLookupMapping[pickupLocationValue];
			else
				log:WarnFormat("PickupLocation mapping not found for {0}", pickupLocationValue);
			end

			CreateRequest(patronId, pickupLocationValue, idList[0].Id, idList[0].Title, false);
		end
	end)

    if (not success) then
		local tn = GetFieldValue("Transaction", "TransactionNumber");
		local errorMessage = "An error occurred when searching ReShare for TN " .. tn .. ": " .. tostring(TraverseError(err));
        log:Error(errorMessage);
		AddNote(tn, errorMessage);
		ExecuteCommand("Route", { tn, Settings.ErrorQueue});
    end
end

function GetItemsAvailableInSharedIndex()
	log:Info("Reshare: Searching Shared Index");
	local idList = nil;

	local oclcNumber = GetFieldValue("Transaction","ESPNumber");
	if (oclcNumber ~= nil and oclcNumber ~= "") then
		idList = searchClient:QueryByOclcNumber(oclcNumber);
	end
	if (idList ~= nil and idList.Count > 0) then
		return idList;
	end

	local isxn = GetFieldValue("Transaction","ISSN");
	if (isxn ~= nil and isxn ~= "") then
		idList = searchClient:QueryByIsxn(isxn);
	end
	if (idList ~= nil and idList.Count > 0) then
		return idList;
	end

	local title = GetFieldValue("Transaction","LoanTitle");
	local author = GetFieldValue("Transaction","LoanAuthor");
	local date = GetFieldValue("Transaction","LoanDate");
	if (title ~= nil and title ~= "") then
		idList = searchClient:QueryByBibInfo(title, author, date);
	end

	if (idList == nil or idList.Count == 0) then
		log:Debug("No results found");
		return idList;
	elseif (idList.Count > 1) then
		log:Debug("Unable to differentiate between multiple records in a bib info search");
		return nil;
	elseif (idList.Count == 1) then
		return idList;
	end

	return idList;
end

function CreateRequest(patronId, pickupLocation, instanceId, title, isDsp)
	log:Debug("ReShare: Creating ReShare Request");

	local tn = GetFieldValue("Transaction","TransactionNumber");

	local volume = GetFieldValue("Transaction", "PhotoJournalVolume");
	if IsNilOrBlank(volume) then
		volume = GetFieldValue("Transaction", "CitedVolume");
		volume = TruncateWithEllipsis(volume, 255);
	end

	local patronNotes = GetPatronNotes(tn);

	local success, result = pcall(function() return requestClient:CreateReshareRequest(patronId, pickupLocation, instanceId, title, tn, volume, patronNotes) end);

	if success then
		--check for nil HRID and add placeholder value
		if (result.Hrid == nil or result.Hrid == "") then
			local message = "A ReShare request was created, but the local ID could not be retrieved. ReShare request id: "..result.Id;
			log:Warn(message);
			ExecuteCommand("AddNote", { tn, message });
			SetFieldValue("Transaction", "ILLNumber", Settings.RequestPlaceholderId);
			SaveDataSource("Transaction");
			ExecuteCommand("Route", { tn, Settings.RequestSentQueue});
		else
			log:Debug("ReShare: Request created. Updating ILL fields.");
			SetFieldValue("Transaction", "ILLNumber", result.Hrid);
			SetFieldValue("Transaction", "SystemID", "Reshare:"..Settings.Tenant);
			SaveDataSource("Transaction");
			ExecuteCommand("Route", { tn, Settings.RequestSentQueue});
			log:Debug("ReShare: ILL fields updated");
		end
    else
		local errorMessage;
		if (result.InnerException:GetType().Name == "ReshareException") then
			errorMessage = "A ReShare request was created, but the local ID could not be retrieved. ReShare request id: "..result.InnerException.RequestId;
			log:Warn(errorMessage);
			AddNote(tn, errorMessage);
			SetFieldValue("Transaction", "ILLNumber", Settings.RequestPlaceholderId);
			SaveDataSource("Transaction");
			ExecuteCommand("Route", { tn, Settings.RequestSentQueue});
		else
			errorMessage = "An error occurred when creating a ReShare request for TN " .. tn .. ": " .. tostring(TraverseError(result));
			log:Error(errorMessage);
			AddNote(tn, errorMessage);

			-- We don't want to route to the normal error queue when running as DSP.
			if isDsp == false then
				ExecuteCommand("Route", { tn, Settings.ErrorQueue});
			end
		end
	end
end

function CheckReshareStatusUpdates()
	log:Info("Checking for ReShare status updates");

	local success, err = pcall(function()
		log:Debug("Updating requests missing HRIDs.");
		UpdateMissingHrids();

		hridList = {};
		updateParams = {};

		local unfilledStateCodes = AtlasHelpers.StringSplit(",", Settings.UnfilledStateCodes);
		for _, stateCode in ipairs(unfilledStateCodes) do
			stateCode = AtlasHelpers.Trim(stateCode);
			if (stateCode ~= nil and stateCode ~= "") then
				GatherReshareUpdateParams(hridList, updateParams, stateCode, Settings.UnfilledQueue, false, false)
			else
				log:Warn("Skipping unfilled check: no statecodes defined in settings.");
			end
		end

		local filledStateCodes = AtlasHelpers.StringSplit(",", Settings.FilledStateCodes);
		for _, stateCode in ipairs(filledStateCodes) do
			stateCode = AtlasHelpers.Trim(stateCode);
			if (stateCode ~= nil and stateCode ~= "") then
				GatherReshareUpdateParams(hridList, updateParams, stateCode, "Request Finished", true, false)
			else
				log:Warn("Skipping completed check: no statecodes defined in settings.");
			end
		end

		if (Settings.ProcessIdleRequests == true) then
			GatherReshareUpdateParams(hridList, updateParams, "REQ_IDLE", Settings.ErrorQueue, false, true)
		end

		PerformReshareUpdates(hridList, updateParams);
	end);

	if (not success) then
		log:Error("An error occurred checking ReShare for status updates: " .. tostring(TraverseError(err)));
	end
end

function GatherReshareUpdateParams(hridList, updateParams, stateCode, newStatus, includeLender, includeAuditMessage)
	log:DebugFormat("Checking for requests in status {0}", stateCode);
	local dateTime = types["System.DateTimeOffset"].FromUnixTimeSeconds(lastPolledTime).DateTime.Date;

	local success, result = pcall(function() return requestClient:GetRequestsByStateCode(stateCode, dateTime) end );

	if (not success) then
		log:Error("An error was encountered retrieving requests by statecode "..stateCode);
		if (result.InnerException and result.InnerException.Message) then
			log:Error(result.InnerException.Message);
		end
		return;
	end

	local reshareRequestList = result;
	log:Debug(reshareRequestList.Results.Count .. " ReShare requests found in state " .. stateCode);
	if (reshareRequestList.Results.Count > 0) then
		local requestEnumerator = reshareRequestList.Results:GetEnumerator();
		while (requestEnumerator:MoveNext()) do
			local request = requestEnumerator.Current;
			if (request == nil) then
				log:Warn("GetRequestsByStateCode returned null response");
			elseif (request.Hrid == nil) then
				log:Warn("GetRequestsByStateCode returned response with null Hrid");
			else

				local params = {};
				params.newStatus = newStatus;

				if (includeLender == true) then
					params.lender = request.ResolvedSupplier.Symbol
				else
					params.lender = nil;
				end;

				if (includeAuditMessage == true and request.audit ~= nil and request.audit[0] ~= nil and request.audit[0].message ~= nil) then
					params.auditMessage = request.audit[0].message;
				else
					params.auditMessage = nil;
				end

				updateParams[request.Hrid] = params;
				table.insert(hridList, request.Hrid);
			end
		end
	end
end

function PerformReshareUpdates(hridList, updateParams)
	if (hridList ~= nil and #hridList > 0) then
		local success, result = pcall(GetTNsFromHRIDs, hridList);
		if (success) then
			ProcessDataContexts("TransactionNumber", result, "UpdateILLiadTransaction");
		else
			log:ErrorFormat("Unable to retrieve TNs from HRID list:"..result.Message);
			if (result.InnerException ~= nil) then
				log:Debug(result.InnerException.Message);
			end
		end
	end
end

function GetTNsFromHRIDs(hridList)
	local tns = {};

	local connection = CreateManagedDatabaseConnection();
	connection:Connect();
	connection.QueryString = "SELECT TransactionNumber FROM Transactions WHERE ILLNumber IN ('" .. table.concat(hridList, "','") .. "') AND TransactionStatus = '"..Settings.RequestSentQueue.."';";
	local transactions = connection:Execute();

	for i = 0, transactions.Rows.Count - 1 do
		table.insert(tns, transactions.Rows:get_Item(i):get_Item(0));
	end
	connection:Dispose();

	return tns;
end

function UpdateILLiadTransaction()
	local tn = GetFieldValue("Transaction", "TransactionNumber");
	local hrid = GetFieldValue("Transaction", "ILLNumber");
	local params = updateParams[hrid];
	local userNvtgc = GetFieldValue("User", "NVTGC");

	if (params == nil) then
		log:DebugFormat("No update parameters found for tn {tn}. Skipping this transaction.", tn);
		return;
	end

	--Confirm that this is a request we should be processing
	if (ILLiad.GetParentNVTGC(userNvtgc)):lower() == string.lower(Settings.Nvtgc):lower() then
		log:Debug("Routing transaction "..tn.." to "..params.newStatus);
		ExecuteCommand("Route", {tn, params.newStatus});

		if (params.lender ~= nil and params.lender ~= "") then
			SetFieldValue("Transaction", "LendingLibrary", params.lender);
		end

		if (params.auditMessage ~= nil and params.auditMessage ~= "") then
			AddNote(tn, params.auditMessage);
		end
		SaveDataSource("Transaction");
	else
		log:Debug("Skipping status update for transaction " .. tn .. " because it is not within this addon's NVTGC.");
	end
end

function UpdateMissingHrids()
	local connection = CreateManagedDatabaseConnection();
	connection:Connect();
	connection.QueryString = "SELECT TransactionNumber, NVTGC FROM Transactions INNER JOIN Users ON Transactions.Username = Users.Username WHERE TransactionStatus = @TransactionStatus AND SystemId = @SystemId AND ILLNumber = @RequestPlaceholderId";
	connection:AddParameter('@TransactionStatus', Settings.RequestSentQueue, types.SqlDbType.NVarChar);
	connection:AddParameter('@SystemId', "Reshare:"..Settings.Tenant, types.SqlDbType.NVarChar);
	connection:AddParameter('@RequestPlaceholderId', Settings.RequestPlaceholderId, types.SqlDbType.NVarChar);
	local transactions = connection:Execute();

	local tns = {};
    for i = 0, transactions.Rows.Count - 1 do
		local tn = transactions.Rows:get_Item(i):get_Item(0);
		local userNvtgc = transactions.Rows:get_Item(i):get_Item(1);

		--Confirm that this is a request we should be processing
		if (ILLiad.GetParentNVTGC(userNvtgc)):lower() == string.lower(Settings.Nvtgc):lower() then
			table.insert(tns, tn);
		else
			log:Debug("Skipping missing HRID update for " .. tn .. " because it is not within this addon's NVTGC.");
		end
    end

	ProcessDataContexts("TransactionNumber", tns, "UpdateMissingHrid");

	connection:Dispose();
end

function UpdateMissingHrid()
	log:Debug("[UpdateMissingHrid]");

	local connection = CreateManagedDatabaseConnection();
	connection:Connect();
	connection.QueryString = "SELECT Note FROM Notes WHERE TransactionNumber = @TN AND NOTE LIKE '%ReShare request id:%'";
	connection:AddParameter('@TN', GetFieldValue("Transaction","TransactionNumber"), types.SqlDbType.Int);
	local notes = connection:Execute();

	local guid = "";
	if (notes.Rows.Count == 1) then
		local note = notes.Rows:get_Item(0):get_Item(0);
		guid = note:match("%x%x%x%x%x%x%x%x%-%x%x%x%x%-%x%x%x%x%-%x%x%x%x%-%x%x%x%x%x%x%x%x%x%x%x%x");
	end

	if (guid ~= nil and guid ~= "") then
		local success, result = pcall(function() return requestClient:GetRequestDetails(guid) end);
		if (success and result.Hrid ~= nil and result.Hrid ~= "") then
			log:Debug("HRID retrieved for TN "..GetFieldValue("Transaction","TransactionNumber")..": "..result.Hrid);
			SetFieldValue("Transaction","ILLNumber", result.Hrid);
			SaveDataSource("Transaction");
		else
			if (result.Message ~= nil and result.Message ~= "") then
				log:Warn("Unable to retrieve HRID for TN "..GetFieldValue("Transaction","TransactionNumber")..": "..result.Message);
			end
		end
	else
		-- If there is no GUID in the notes, there's no point in checking this request again later, so we'll remove the ILLNumber.
		local tn = GetFieldValue("Transaction","TransactionNumber");
		local message = "Unable to retrieve HRID for TN "..GetFieldValue("Transaction","TransactionNumber")..": no GUID present in notes.";
		log:Warn(message);
		AddNote(tn, message);
		ExecuteCommand("Route",{tn, Settings.ErrorQueue});
		SetFieldValue("Transaction","ILLNumber","");
		SaveDataSource("Transaction");
	end

	connection:Dispose();
end

function GetPatronNotes(tn)
	local connection = CreateManagedDatabaseConnection();
	connection:Connect();
	connection.QueryString = "SELECT Note FROM Notes WHERE TransactionNumber = '" .. tn .. "' AND NoteType = 'User' ORDER BY NoteDate DESC;";
	local results = connection:Execute();

	local notesArray = {};
    for i = 0, results.Rows.Count - 1 do
		log:Debug("Note " .. i+1 .. ": " .. tostring(results.Rows:get_item(i):get_Item(0)));
		notesArray[#notesArray+1] = results.Rows:get_item(i):get_Item(0);
    end

	connection:Dispose();
	
	local patronNotes = table.concat(notesArray, "; ");
	patronNotes = TruncateWithEllipsis(patronNotes, 255);
	
	return patronNotes;
end

function BuildPickupLocationMappings()
	local mappings = AtlasHelpers.StringSplit(",", GetSetting("PickupLocationLookupMapping"));
	for i = 1, #mappings do
		local keyValue = {};

		log:DebugFormat("Processing PickupLocation Mapping {0}", mappings[i]);

		local equalPosition = string.find(mappings[i], "=");

		if (equalPosition == nil) then
			log:ErrorFormat("Invalid Location mapping found: {0}",  mappings[i]);
			ExecuteCommand("WriteSystemLog", "ReShare: Invalid Location mapping found: " ..  mappings[i]);
		else
			local illiadLocation = AtlasHelpers.Trim(string.sub(mappings[i], 0, equalPosition - 1));
			local reshareLocation = AtlasHelpers.Trim(string.sub(mappings[i], equalPosition + 1));
			log:DebugFormat("ILLiad Location: {0}. ReShare Location: {1}", illiadLocation, reshareLocation);
			Settings.PickupLocationLookupMapping[illiadLocation] = reshareLocation;
		end
	end
end

function AddNote(tn, note)
	ExecuteCommand("AddNote", {tn, Settings.NoteAndLogPrefix .. " - " .. note});
end

function IsNilOrBlank(value)
	if not value or value == "" then
		return true;
	else
		return false;
	end
end

function TruncateWithEllipsis(str, characterLimit)

	if str:len() > characterLimit then
		-- Truncate with an ellipsis so users know the string has been truncated.
		str = str:sub(1, characterLimit):gsub("...$", "...");
	end

	return str;
end

function TraverseError(e)
    if not e.GetType then
        -- Not a .NET type
        return nil;
    else
        if not e.Message then
            -- Not a .NET exception
            log:Debug(e:ToString());
            return nil;
        end
    end

    log:Debug(e.Message);

    if e.InnerException then
        return TraverseError(e.InnerException);
    else
        return e.Message;
    end
end